#include "lidar_processing.h"
#include <pcl/filters/radius_outlier_removal.h>
#include <pcl/filters/voxel_grid.h>

#include <pcl/io/pcd_io.h>
#include <chrono>

#include <pcl/filters/statistical_outlier_removal.h>
#include <pcl/filters/crop_box.h>
#include <pcl/point_types.h>
#include <pcl/common/common.h>
#include <iostream>
#include <pcl/filters/passthrough.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/common/common_headers.h>
#include <cmath>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/common/centroid.h>
#include <pcl/search/kdtree.h>
#include <vector>
#include <pcl/point_cloud.h>
//#include <pcl/visualization/pcl_visualizer.h>
//#include <pcl/visualization/point_picking_event.h>
//#include <pcl/visualization/keyboard_event.h>
#include <pcl/io/ply_io.h>
#include <pcl/segmentation/region_growing.h>
//#include <opencv2/opencv.hpp>
#include <thread>
#include <omp.h>
#include <pcl/kdtree/kdtree_flann.h>
//#include "my_proto.pb.h"
#include <pcl/filters/passthrough.h>
#include <pcl/filters/conditional_removal.h>
#include <pcl/filters/radius_outlier_removal.h>


namespace Lidar_processing {

    pcl::PointCloud<pcl::PointXYZ>::Ptr RadiusOutlierRemoval(
            pcl::PointCloud<pcl::PointXYZ>::Ptr input_cloud,  
            float radius, int min_neighbors) {
        pcl::PointCloud<pcl::PointXYZ>::Ptr filtered_cloud(new pcl::PointCloud<pcl::PointXYZ>());
        pcl::RadiusOutlierRemoval<pcl::PointXYZ> rad_out; 
        rad_out.setInputCloud(input_cloud);
        rad_out.setRadiusSearch(radius);
        rad_out.setMinNeighborsInRadius(min_neighbors);
        rad_out.filter(*filtered_cloud);
        return filtered_cloud; 
    }

    pcl::PointCloud<pcl::PointXYZ>::Ptr VoxelGridFilter(
            pcl::PointCloud<pcl::PointXYZ>::Ptr input_cloud, 
            float leaf_size) {
        pcl::PointCloud<pcl::PointXYZ>::Ptr filtered_cloud(new pcl::PointCloud<pcl::PointXYZ>());
        pcl::VoxelGrid<pcl::PointXYZ> voxel_filter;
        voxel_filter.setInputCloud(input_cloud);
        voxel_filter.setLeafSize(leaf_size, leaf_size, leaf_size);
        voxel_filter.filter(*filtered_cloud);
        return filtered_cloud;
    }

    pcl::PointCloud<pcl::PointXYZ>::Ptr filterRange(
            const pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_in,
            const Eigen::Vector4f& min_range, const Eigen::Vector4f& max_range) {
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::CropBox<pcl::PointXYZ> box_filter;
    box_filter.setInputCloud(cloud_in);
    box_filter.setMin(min_range);
    box_filter.setMax(max_range);
    box_filter.filter(*cloud_filtered);

    return cloud_filtered;
}

    pcl::PointCloud<pcl::PointXYZ>::Ptr filterGround(
            const pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_in, float distance_threshold)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered(new pcl::PointCloud<pcl::PointXYZ>);
    
    pcl::SACSegmentation<pcl::PointXYZ> seg;
    pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
    pcl::ModelCoefficients::Ptr coefficients(new pcl::ModelCoefficients);

    seg.setOptimizeCoefficients(true);
    seg.setModelType(pcl::SACMODEL_PLANE);
    seg.setMethodType(pcl::SAC_RANSAC);
    seg.setDistanceThreshold(distance_threshold);

    seg.setInputCloud(cloud_in);
    seg.segment(*inliers, *coefficients);

    pcl::ExtractIndices<pcl::PointXYZ> extract;
    extract.setInputCloud(cloud_in);
    extract.setIndices(inliers);
    extract.setNegative(true); // Extract non-ground points
    extract.filter(*cloud_filtered);

    return cloud_filtered;
}

    void getMinMaxParallel(
            const pcl::PointCloud<pcl::PointXYZ>& cloud, pcl::PointXYZ& min_pt, pcl::PointXYZ& max_pt) {
    const size_t size = cloud.size();
    const size_t num_threads = std::thread::hardware_concurrency();
    const size_t chunk_size = (size + num_threads - 1) / num_threads;  // Round up

    std::vector<std::thread> threads;
    std::vector<pcl::PointXYZ> min_pts(num_threads, pcl::PointXYZ(std::numeric_limits<float>::max(), std::numeric_limits<float>::max(), std::numeric_limits<float>::max()));
    std::vector<pcl::PointXYZ> max_pts(num_threads, pcl::PointXYZ(std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest()));

    // Function to update min-max in a chunk
    auto updateMinMax = [&](size_t start, size_t end, size_t thread_id) {
        for (size_t i = start; i < end && i < size; ++i) {
            min_pts[thread_id].x = std::min(min_pts[thread_id].x, cloud[i].x);
            min_pts[thread_id].y = std::min(min_pts[thread_id].y, cloud[i].y);
            min_pts[thread_id].z = std::min(min_pts[thread_id].z, cloud[i].z);

            max_pts[thread_id].x = std::max(max_pts[thread_id].x, cloud[i].x);
            max_pts[thread_id].y = std::max(max_pts[thread_id].y, cloud[i].y);
            max_pts[thread_id].z = std::max(max_pts[thread_id].z, cloud[i].z);
        }
    };

    // Launch threads
    for (size_t i = 0; i < num_threads; ++i) {
        size_t start = i * chunk_size;
        size_t end = std::min((i + 1) * chunk_size, size);
        threads.emplace_back(updateMinMax, start, end, i);
    }

    // Join threads
    for (auto& thread : threads) {
        thread.join();
    }

    // Combine results from all threads using a parallel reduction
    min_pt = pcl::PointXYZ(std::numeric_limits<float>::max(), std::numeric_limits<float>::max(), std::numeric_limits<float>::max());
    max_pt = pcl::PointXYZ(std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest());

    for (size_t i = 0; i < num_threads; ++i) {
        min_pt.x = std::min(min_pt.x, min_pts[i].x);
        min_pt.y = std::min(min_pt.y, min_pts[i].y);
        min_pt.z = std::min(min_pt.z, min_pts[i].z);

        max_pt.x = std::max(max_pt.x, max_pts[i].x);
        max_pt.y = std::max(max_pt.y, max_pts[i].y);
        max_pt.z = std::max(max_pt.z, max_pts[i].z);
    }
}


std::vector<LidarPerceptionOutput> euclideanClustering(const pcl::PointCloud<pcl::PointXYZ>::Ptr& cloud) {
    // Create Euclidean Cluster Extraction object
    auto start = std::chrono::high_resolution_clock::now();
    pcl::EuclideanClusterExtraction<pcl::PointXYZ> ec;

    // Create a KD-Tree
    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);

    // Set the KD-Tree as the search method for the Euclidean Cluster Extraction
    tree->setInputCloud(cloud);
    ec.setSearchMethod(tree);

    // Set input cloud
    ec.setInputCloud(cloud);

    // Set cluster parameters
    ec.setClusterTolerance(0.3); // Set the distance threshold
    ec.setMinClusterSize(20);      // Minimum number of points in a cluster
    ec.setMaxClusterSize(5000);  // Maximum number of points in a cluster

    // Perform clustering
    std::vector<pcl::PointIndices> cluster_indices;
    ec.extract(cluster_indices);
	
    // Stop timer
    auto stop = std::chrono::high_resolution_clock::now();   
    // Calculate duration 
    auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(stop - start);
    // Print running time
    std::cout << "Clustering took " << duration.count() << " milliseconds" << std::endl;
	
	/*
	 // Iterate over each cluster to compute and print bounding boxes
    int cluster_id = 1;
    for (const auto& indices : cluster_indices) {
        pcl::PointCloud<pcl::PointXYZ>::Ptr cluster(new pcl::PointCloud<pcl::PointXYZ>);
        for (const auto& index : indices.indices) {
            cluster->points.push_back(cloud->points[index]);
        }

        // Compute bounding box
        pcl::PointXYZ min_point, max_point;
        pcl::getMinMax3D(*cluster, min_point, max_point);

        // Print bounding box coordinates
        std::cout << "Bounding Box " << cluster_id << ": Min (" 
                  << min_point.x << ", " << min_point.y << ", " << min_point.z 
                  << "), Max (" << max_point.x << ", " << max_point.y << ", " << max_point.z << ")" << std::endl;

        cluster_id++;
    }*/

     //std::vector<BoundingBox> outputbboxes;
    std::vector<LidarPerceptionOutput> outputbboxes; 

    

    int cluster_id1 = 1;
    for (const auto& indices : cluster_indices) {
        // ... (code to compute bounding box)
         pcl::PointCloud<pcl::PointXYZ>::Ptr cluster(new pcl::PointCloud<pcl::PointXYZ>);
        for (const auto& index : indices.indices) {
            cluster->points.push_back(cloud->points[index]);
        }
        pcl::PointXYZ min_point, max_point;
        pcl::getMinMax3D(*cluster, min_point, max_point);
        
        // Store bounding box values in the vector
        LidarPerceptionOutput output;
        output.bbox.xmin = min_point.x;
        output.bbox.ymin = min_point.y;
        //bbox.minz = min_point.z;
        output.bbox.xmax = max_point.x;
        output.bbox.ymax = max_point.y;
        //bbox.maxz = max_point.z;

        output.bbox3d.minx = min_point.x;
        output.bbox3d.miny = min_point.y;
        output.bbox3d.minz = min_point.z;
        output.bbox3d.maxx = max_point.x;
        output.bbox3d.maxy = max_point.y;
        output.bbox3d.maxz = max_point.z;


        output.confidence = 1;
        //output.label = 100;

        //Fill position
        output.position.x = (output.bbox.xmin + output.bbox.xmax) / 2;
        output.position.y = (output.bbox.ymin + output.bbox.ymax) / 2;
        output.position.z = (output.bbox.xmin + output.bbox.xmax) / 2;

        output.length = output.bbox.xmax - output.bbox.xmin;
        output.width = output.bbox.ymax - output.bbox.ymin;
        output.height = max_point.z - min_point.z; //zmax - zmin

        // Fill polygon_point directly from bboxes.data
        // Front face vertices
        output.polygon_point.push_back({min_point.x, min_point.y, min_point.z});
        output.polygon_point.push_back({max_point.x, min_point.y, min_point.z});
        output.polygon_point.push_back({max_point.x, max_point.y, min_point.z});
        output.polygon_point.push_back({min_point.x, max_point.y, min_point.z});

        // Back face vertices
        output.polygon_point.push_back({min_point.x, min_point.y, max_point.z});
        output.polygon_point.push_back({max_point.x, min_point.y, max_point.z});
        output.polygon_point.push_back({max_point.x, max_point.y, max_point.z});
        output.polygon_point.push_back({min_point.x, max_point.y, max_point.z});



        outputbboxes.push_back(output);

        cluster_id1++;
    }
    

    return outputbboxes;


   
	
	
    /*
    // Create a PCL Visualizer
    
    pcl::visualization::PCLVisualizer::Ptr viewer(new pcl::visualization::PCLVisualizer("Cluster viewer"));

    // Set background color
    viewer->setBackgroundColor(0, 0, 0);

    // Add the original point cloud to the viewer
    pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZ> single_color(cloud, 0, 255, 0);
    viewer->addPointCloud<pcl::PointXYZ>(cloud, single_color, "original_cloud");
    viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 1, "original_cloud");

    // Visualize clusters with bounding boxes
    int cluster_color = 1;
    for (const auto& indices : cluster_indices) {
        pcl::PointCloud<pcl::PointXYZ>::Ptr cluster(new pcl::PointCloud<pcl::PointXYZ>);
        for (const auto& index : indices.indices) {
            cluster->points.push_back(cloud->points[index]);
        }

        // Compute bounding box
        pcl::PointXYZ min_point, max_point;
        pcl::getMinMax3D(*cluster, min_point, max_point);

        // Add bounding box
        viewer->addCube(min_point.x, max_point.x, min_point.y, max_point.y, min_point.z, max_point.z,
                        cluster_color * 0.1, 0, 0, "box_" + std::to_string(cluster_color));

        cluster_color++;
    }

    // Spin the viewer
    while (!viewer->wasStopped()) {
        viewer->spinOnce(100);
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }*/
}

pcl::PointCloud<pcl::PointXYZ>::Ptr filterRoadBoundary(const pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_in, float min_boundary, float max_boundary)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PassThrough<pcl::PointXYZ> pass;
    pass.setInputCloud(cloud_in);
    pass.setFilterFieldName("y"); // Assuming the road boundary is aligned with the y-axis
    pass.setFilterLimits(min_boundary, max_boundary);
    pass.filter(*cloud_filtered);

    return cloud_filtered;
}


// Conversion function from PointCloud message to PCL point cloud
// pcl::PointCloud<pcl::PointXYZ>::Ptr pointCloudToPcl(const PointCloud& pointCloudMsg) {
//     pcl::PointCloud<pcl::PointXYZ>::Ptr pclCloud(new pcl::PointCloud<pcl::PointXYZ>);

//     for (const auto& point : pointCloudMsg.points()) {
//         pcl::PointXYZ pclPoint;
//         pclPoint.x = point.x();
//         pclPoint.y = point.y();
//         pclPoint.z = point.z();
//         pclCloud->push_back(pclPoint);
//     }

//     return pclCloud;
// }



// Function to apply transformation matrix to a point cloud
pcl::PointCloud<pcl::PointXYZ>::Ptr transformPointCloud(const pcl::PointCloud<pcl::PointXYZ>::Ptr& cloud, const Eigen::Matrix4f& transformation_matrix) {
    pcl::PointCloud<pcl::PointXYZ>::Ptr transformed_cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::transformPointCloud(*cloud, *transformed_cloud, transformation_matrix);
    return transformed_cloud;
}



// Function to stitch multiple point clouds together
pcl::PointCloud<pcl::PointXYZ>::Ptr stitchPointClouds(const std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr>& clouds) {
    pcl::PointCloud<pcl::PointXYZ>::Ptr stitched_cloud(new pcl::PointCloud<pcl::PointXYZ>);

    // Concatenate point clouds
    for (const auto& cloud : clouds) {
        *stitched_cloud += *cloud;
    }

    return stitched_cloud;
}


// Function for range filtering using pcl::PassThrough
pcl::PointCloud<pcl::PointXYZ>::Ptr rangeFilter(const pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_in, double xmin, double ymin, double zmin,
                                      double xmax, double ymax, double zmax) {
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PassThrough<pcl::PointXYZ> pass;
    
    pass.setInputCloud(cloud_in);
    pass.setFilterFieldName("x");
    pass.setFilterLimits(xmin, xmax);
    pass.filter(*cloud_filtered);

    pass.setInputCloud(cloud_in);
    pass.setFilterFieldName("y");
    pass.setFilterLimits(ymin, ymax);
    pass.filter(*cloud_filtered);

    pass.setInputCloud(cloud_in);
    pass.setFilterFieldName("z");
    pass.setFilterLimits(zmin, zmax);
    pass.filter(*cloud_filtered);

    return cloud_filtered;
}

// Function for noise filtering (removing zero points and duplicates)
pcl::PointCloud<pcl::PointXYZ>::Ptr noiseFilter(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud) {
    // Remove zero points
     // Create ConditionAnd object
    pcl::ConditionAnd<pcl::PointXYZ>::Ptr zero_cond (new pcl::ConditionAnd<pcl::PointXYZ>());

    // Add conditions
    zero_cond->addComparison(pcl::FieldComparison<pcl::PointXYZ>::ConstPtr (
            new pcl::FieldComparison<pcl::PointXYZ>("x", pcl::ComparisonOps::EQ, 0.0)));

    zero_cond->addComparison(pcl::FieldComparison<pcl::PointXYZ>::ConstPtr (  
            new pcl::FieldComparison<pcl::PointXYZ>("y", pcl::ComparisonOps::EQ, 0.0)));

    zero_cond->addComparison(pcl::FieldComparison<pcl::PointXYZ>::ConstPtr (
            new pcl::FieldComparison<pcl::PointXYZ>("z", pcl::ComparisonOps::EQ, 0.0)));

    // Create filter 
    pcl::ConditionalRemoval<pcl::PointXYZ> zero_cond_removal;

    // Set condition
    zero_cond_removal.setCondition(zero_cond);

    // Set other parameters
    zero_cond_removal.setInputCloud(cloud);
    zero_cond_removal.setKeepOrganized(true);

    // Apply filter
    zero_cond_removal.filter(*cloud);

    // Remove duplicate points
    pcl::RadiusOutlierRemoval<pcl::PointXYZ> radius_outlier_removal;
    radius_outlier_removal.setInputCloud(cloud);
    radius_outlier_removal.setRadiusSearch(0.01); 
    radius_outlier_removal.setMinNeighborsInRadius(5); 
    radius_outlier_removal.setKeepOrganized(true); 
    radius_outlier_removal.filter(*cloud);

    return cloud;
}

pcl::PointCloud<pcl::PointXYZ>::Ptr convertToPCL(const LidarPerceptionInput& input) {
    pcl::PointCloud<pcl::PointXYZ>::Ptr pclCloudPtr(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>& pclCloud = *pclCloudPtr; // Reference to the point cloud
    pclCloud.width = input.points.size();
    pclCloud.height = 1;
    pclCloud.is_dense = false;
    pclCloud.points.resize(pclCloud.width * pclCloud.height);

    for (size_t i = 0; i < input.points.size(); ++i) {
        pclCloud.points[i].x = input.points[i].x;
        pclCloud.points[i].y = input.points[i].y;
        pclCloud.points[i].z = input.points[i].z;
    }

    return pclCloudPtr;
}


    

} // namespace lidar_processing
